#      ******************************************************************
#      *                                                                *
#      *                           DPi Network                          *
#      *                                                                *
#      *            Stan Reifel                     8/19/2022           *
#      *                                                                *
#      ******************************************************************

#
# Protocol for RS485 serial communication between the Master and a Slave module:
# 
# Commands are sent from the Master to a Slave.  The slave on the RS485 network is
# identified by the Slave's address.  Each Slave must have a unique 6 bit SlaveAddress.
#
# Commands from the Master include a CommandByte and up to 15 bytes of additional 
# data.  The Slave must acknowledge that it received the Master's command and that 
# the data it received is intact.  The Master can send commands to have the Slave 
# do things, or it can send commands requesting data from the Slave.
#
# When the Slave receives a command from the Master, it must acknowledge the command.
# This acknowledgement is used to let the Master know the command was received
# successfully.  When the Master sends a command requesting data from the Slave, the 
# Slave sends the data by adding it to the acknowledgement (which can include up
# to 15 bytes of data from the Slave).
#
# After the Master sends a command, it waits for an acknowledgement from the Slave.
# If the Master doesn't receive the acknowledgement within a few milliseconds, it
# resends the command (up to n times).  If the Slave receives a command with corrupted 
# data, the Slave simply ignores the command by not executing it or sending any 
# acknowledgement (causing the Master to resend). 
# 
# All data transfers are initiated by the Master.  The Slave can not asynchronously
# transmit data that wasn't requested by a command from the Master.
#
# Byte format of command packets sent from the Master to Slave:
#   * HeaderByte: Always 0xaa
#   * SlaveAddress: 6 bits, note bit 7 must be 1, bit 6 must be 0. 
#   * CommandByte
#   * DataCount: Count of bytes in the DataField (0 - 15):
#      - The value's low nibble = DataCount
#      - The value's high nibble = (~DataCount) << 4   (one's complment of the DataCount)
#   * DataField: Between 0 and 15 bytes of additional data sent by the Master
#   * Checksum: The 8 bit sum of all bytes in the packet including the HeaderByte, 
#              DataCount value, SlaveAddress value, CommandByte, and all bytes in 
#              the DataField.
#
# There are two packet formats the Slave uses when acknowledging the Master's commands.
# If the Master isn't requesting data from the Slave, the acknowledgement format is:
#   * HeaderByte: 0x77
#   * Checksum: (will always be 0x77)
# If the Master is requesting data from the Slave, the format is:
#   * HeaderByte: 0x78
#   * DataCount: Count of bytes in the DataField (1 - 15):
#      - The value's low nibble = DataCount
#      - The value's high nibble = (~DataCount) << 4   (one's complment of the DataCount)
#  * DataField: Between 1 and 15 bytes of data sent from the Slave
#  * Checksum: The 8 bit sum of all bytes in the acknowledgement packet including the 
#              HeaderByte, DataCount value, and all bytes in the DataField.
#

from time import time
import serial
serialPort = serial


#
# Master/Slave communication constants
#
_RS485_BAUDRATE = 115200
_RS485_RETRY_ATTEMPTS = 3
_COMMAND_PACKET_HEADER = 0xaa
_ACKNOWLEDGEMENT_PACKET_HEADER_WITH_NO_DATA = 0x77
_ACKNOWLEDGEMENT_PACKET_HEADER_WITH_DATA = 0x78

#
# state values for: receivePacketState
#
_RECEIVE_PACKET_STATE__WAITING_FOR_HEADER = 0
_RECEIVE_PACKET_STATE__WAITING_FOR_DATA_COUNT = 1
_RECEIVE_PACKET_STATE__WAITING_DATA = 2
_RECEIVE_PACKET_STATE__WAITING_FOR_CHECKSUM = 8

#
# vars global to this module
#
_dataTransmitBuffer = bytearray(16)
_dataTransmitIndex = 0
_dataReceiveBuffer = bytearray(16)
_dataReceiveIndex = 0

class DPiNetwork:
    #
    # initialize the RS485 serial communication
    #
    def __init__(self):
        global serialPort
        global _dataTransmitIndex
        global _dataReceiveIndex

        serialPort = serial.Serial(
            port="/dev/serial0", 
            baudrate = _RS485_BAUDRATE,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            bytesize=serial.EIGHTBITS)

        _dataTransmitIndex = 0
        _dataReceiveIndex = 0


    #
    # send a command to a slave module via a RS485 network, command's additional data must have 
    # already been "Pushed".  After this function returns data from the slave is retrieved by "Popping"
    #    Enter:  slaveAddress = address of slave module (0 - 0x3f)
    #            command = command byte
    #            timeoutPeriodMS = number of milliseconds to wait for the slave to respond
    #    Exit:   [0]: True returned on success, else False
    #            [1]: number of failed packets (0 = no failed packets)
    #
    def sendCommand(self, slaveAddress: int, command: int, timeoutPeriodMS: int):
        global serialPort
        global _dataTransmitIndex
        global _dataTransmitBuffer
        global _dataReceiveIndex
        global _dataReceiveBuffer
        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER
        timeoutPeriodSec = timeoutPeriodMS / 1000.0

        # 
        # flush the buffer of data received from the slaves, to make sure it's empty before sending the packet
        #
        serialPort.reset_input_buffer()

        #
        # setup a loop to send packet to slave, retrying if it fails
        #
        sendDataSize = _dataTransmitIndex
        _dataTransmitIndex = 0
        sendDataCountValue = (sendDataSize & 0x0f) + (((~sendDataSize) & 0x0f) << 4)
        slaveAddressValue = (slaveAddress & 0x3f) | 0x80

        for retryCount in range(0, _RS485_RETRY_ATTEMPTS):
            #
            # send packet to slave
            #
            packet = bytearray(5 + sendDataSize)
            checksum = 0
            
            packet[0] = _COMMAND_PACKET_HEADER
            checksum += _COMMAND_PACKET_HEADER
            
            packet[1] = slaveAddressValue
            checksum += slaveAddressValue
            
            packet[2] = command
            checksum += command
            
            packet[3] = sendDataCountValue
            checksum += sendDataCountValue
            
            for i in range(0, sendDataSize):
                packet[4 + i] = _dataTransmitBuffer[i]
                checksum += _dataTransmitBuffer[i]

            packet[4 + sendDataSize] = checksum % 0x100

            serialPort.write(packet)

            #
            # get acknowledgement packet from slave, timing out if it isn't received (or received corrupted)
            #
            startTime = time()
            receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER
            receivedDataSize = 0
            checksum = 0
            
            while(True):
                #
                # check if timed out while waiting for slave's response
                #
                elapsedTime = time() - startTime
                if (elapsedTime > timeoutPeriodSec):
                    break                                  # timed out, exit "while loop" causing packet to be resent to slave 

                #
                # check if there's a new byte of data from the slave
                #
                if serialPort.inWaiting() == 0:
                    continue
                
                c = serialPort.read()[0]
                if receivePacketState == _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER:
                    checksum += c;
                    if c == _ACKNOWLEDGEMENT_PACKET_HEADER_WITH_NO_DATA:
                        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_CHECKSUM;

                    if c == _ACKNOWLEDGEMENT_PACKET_HEADER_WITH_DATA:
                        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_DATA_COUNT;
                
                
                elif receivePacketState == _RECEIVE_PACKET_STATE__WAITING_FOR_DATA_COUNT:
                    checksum += c
                    receivedDataSize = c & 0x0f
                    bytesReceivedCount = 0
                    _dataReceiveIndex = 0

                    if receivedDataSize == 0:
                        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER
                    elif receivedDataSize != (((~c) >> 4) & 0x0f):
                        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER
                    else:
                        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_DATA


                elif receivePacketState == _RECEIVE_PACKET_STATE__WAITING_DATA:
                    checksum += c
                    _dataReceiveBuffer[bytesReceivedCount] = c
                    bytesReceivedCount += 1

                    if bytesReceivedCount == receivedDataSize:
                        receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_CHECKSUM;


                elif receivePacketState == _RECEIVE_PACKET_STATE__WAITING_FOR_CHECKSUM:
                    if c == checksum % 0x100:
                        return True, retryCount


                    receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER


                else:
                    receivePacketState = _RECEIVE_PACKET_STATE__WAITING_FOR_HEADER
        
        return False, retryCount+1


    #
    # push data to the transmit buffer
    #
    def pushUint8(self, value: int):
        global _dataTransmitIndex
        global _dataTransmitBuffer
        _dataTransmitBuffer[_dataTransmitIndex] = value & 0xff;
        _dataTransmitIndex += 1

    def pushInt8(self, value: int):
        global _dataTransmitIndex
        global _dataTransmitBuffer
        _dataTransmitBuffer[_dataTransmitIndex] = value & 0xff;
        _dataTransmitIndex += 1

    def pushUint16(self, value: int):
        global _dataTransmitIndex
        global _dataTransmitBuffer
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 8) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = value & 0xff;
        _dataTransmitIndex += 1

    def pushInt16(self, value: int):
        global _dataTransmitIndex
        global _dataTransmitBuffer
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 8) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = value & 0xff;
        _dataTransmitIndex += 1

    def pushUint32(self, value: int):
        global _dataTransmitIndex
        global _dataTransmitBuffer
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 24) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 16) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 8) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = value & 0xff;
        _dataTransmitIndex += 1

    def pushInt32(self, value: int):
        global _dataTransmitIndex
        global _dataTransmitBuffer
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 24) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 16) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = (value >> 8) & 0xff;
        _dataTransmitIndex += 1
        _dataTransmitBuffer[_dataTransmitIndex] = value & 0xff;
        _dataTransmitIndex += 1


    #
    # get data from the receive buffer
    #
    def popUint8(self) -> int:
        global _dataReceiveIndex
        global _dataReceiveBuffer
        value = _dataReceiveBuffer[_dataReceiveIndex]
        _dataReceiveIndex += 1
        return value

    def popInt8(self) -> int:
        global _dataReceiveIndex
        global _dataReceiveBuffer
        value = _dataReceiveBuffer[_dataReceiveIndex]
        if value & 0x80 > 0:
            value -= 0x100
        _dataReceiveIndex += 1
        return value

    def popUint16(self) -> int:
        global _dataReceiveIndex
        global _dataReceiveBuffer
        value = ((_dataReceiveBuffer[_dataReceiveIndex] << 8) + _dataReceiveBuffer[_dataReceiveIndex + 1])
        _dataReceiveIndex += 2
        return value

    def popInt16(self) -> int:
        global _dataReceiveIndex
        global _dataReceiveBuffer
        value = ((_dataReceiveBuffer[_dataReceiveIndex] << 8) + _dataReceiveBuffer[_dataReceiveIndex + 1])
        if value & 0x8000 > 0:
            value -= 0x10000
        _dataReceiveIndex += 2
        return value

    def popUint32(self) -> int:
        global _dataReceiveIndex
        global _dataReceiveBuffer
        value = ((_dataReceiveBuffer[_dataReceiveIndex] << 24) + 
                (_dataReceiveBuffer[_dataReceiveIndex+1] << 16) + 
                (_dataReceiveBuffer[_dataReceiveIndex+2] << 8) + 
                (_dataReceiveBuffer[_dataReceiveIndex+3]))
        _dataReceiveIndex += 4
        return value

    def popInt32(self) -> int:
        global _dataReceiveIndex
        global _dataReceiveBuffer
        value = ((_dataReceiveBuffer[_dataReceiveIndex] << 24) + 
                (_dataReceiveBuffer[_dataReceiveIndex+1] << 16) + 
                (_dataReceiveBuffer[_dataReceiveIndex+2] << 8) + 
                (_dataReceiveBuffer[_dataReceiveIndex+3]))
        if value & 0x80000000 > 0:
            value -= 0x100000000
        _dataReceiveIndex += 4
        return value



#      ******************************************************************
#      *                                                                *
#      *                        DPiFuncGen Libary                       *
#      *                                                                *
#      *            Stan Reifel                     8/21/2022           *
#      *                                                                *
#      ******************************************************************

from DPiNetwork import DPiNetwork
dpiNetwork = DPiNetwork()


#
# DPiNetwork DPiFuncGen commands  
#
_CMD_DPi_FUNCGEN__PING                  = 0x00
_CMD_DPi_FUNCGEN__INITIALIZE            = 0x01
_CMD_DPi_FUNCGEN__SET_FREQENCY_AND_WAVE = 0x02
_CMD_DPi_FUNCGEN__SET_VOLUME            = 0x03
_CMD_DPi_FUNCGEN__MUTE                  = 0x04


#
# other constants used by this class
#
_DPiNETWORK_TIMEOUT_PERIOD_MS = 5
_DPiNETWORK_BASE_ADDRESS = 0x18



class DPiFuncGen:
    #
    # audio wave shapes
    #
    SINE_WAVE       = 0
    TRIANGLE_WAVE   = 1
    SQUARE_WAVE     = 2

    #
    # attributes local to this class
    #
    _slaveAddress = _DPiNETWORK_BASE_ADDRESS
    _commErrorCount = 0


    #
    # constructor for the DPiFuncGen class
    #
    def __init__(self):
        pass
       

    # ---------------------------------------------------------------------------------
    #                                 Private functions 
    # ---------------------------------------------------------------------------------

    #
    # send a command to the DPiFuncGen, command's additional data must have already been "Pushed". 
    # After this function returns data from the device is retrieved by "Popping"
    #    Enter:  command = command byte
    #    Exit:   True returned on success, else False
    #
    def __sendCommand(self, command: int):
        (results, failedCount) = dpiNetwork.sendCommand(self._slaveAddress, command, _DPiNETWORK_TIMEOUT_PERIOD_MS)
        self._commErrorCount += failedCount;
        return results

 
    # ---------------------------------------------------------------------------------
    #                                Public functions 
    # ---------------------------------------------------------------------------------

    #
    # set the DPiFuncGen board number
    #    Enter:  boardNumber = DPiFuncGen board number (0 - 3)
    #
    def setBoardNumber(self, boardNumber: int):
        self._slaveAddress = _DPiNETWORK_BASE_ADDRESS + boardNumber


    #
    # ping the board
    #    Exit:   True returned on success, else False
    #
    def ping(self):
        return self.__sendCommand(_CMD_DPi_FUNCGEN__PING)


    #
    # initialize the board to its "power on" configuration
    #    Exit:   True returned on success, else False
    #
    def initialize(self):
        return self.__sendCommand(_CMD_DPi_FUNCGEN__INITIALIZE)


    #
    # set the function generator's frequency and waveshape 
    #    Enter:  frequency = cycles per second (10Hz - 25000Hz)
    #            waveShape = AD9833_SINE_WAVE, AD9833_TRIANGLE_WAVE or AD9833_SQUARE_WAVE
    #    Exit:   True returned on success, else False
    #
    def setFrequency(self, frequency: int, waveShape: int):
        if (frequency < 10) or (frequency > 25000):
            return False

        if not ((waveShape == self.SINE_WAVE) or (waveShape == self.TRIANGLE_WAVE) or (waveShape == self.SQUARE_WAVE)):
            return False

        dpiNetwork.pushUint16(frequency)
        dpiNetwork.pushUint8(waveShape)
        return self.__sendCommand(_CMD_DPi_FUNCGEN__SET_FREQENCY_AND_WAVE)


    #
    # set the function generator's volume
    #    Enter:  volume = volume (0 - 63)
    #    Exit:   True returned on success, else False
    #
    def setVolume(self, volume: int):
        if (volume < 0) or (volume > 63):
            return False

        dpiNetwork.pushUint8(volume)
        return self.__sendCommand(_CMD_DPi_FUNCGEN__SET_VOLUME)


    #
    # mute/un-mute the function generator
    #    Enter:  muteFlg = True to mute, False to un-mute
    #    Exit:   True returned on success, else False
    #
    def mute(self, muteFlg: bool):
        dpiNetwork.pushUint8(muteFlg)
        return self.__sendCommand(_CMD_DPi_FUNCGEN__MUTE)
 

    #
    # get the count of communication errors
    #    Exit:   0 return if no errors, else count of errors returned
    #
    def getCommErrorCount(self):
        return self._commErrorCount



#      ******************************************************************
#      *                                                                *
#      *                         DPiRobot Libary                        *
#      *                                                                *
#      *            Stan Reifel                     8/24/2022           *
#      *                                                                *
#      ******************************************************************

from DPiNetwork import DPiNetwork
dpiNetwork = DPiNetwork()


#
# DPiNetwork DPiRobot commands  
#
_CMD_DPiROBOT__PING                                            = 0x00   
_CMD_DPiROBOT__INITIALIZE                                      = 0x01 
_CMD_DPiROBOT__GET_ROBOT_STATE                                 = 0x02    # robot_GetRobotState
_CMD_DPiROBOT__DISABLE_ROBOT_MOTORS                            = 0x03    # enable / disable the stepper motors
_CMD_DPiROBOT__HOME_ROBOT                                      = 0x04    # robot_prepareForMotion / HomeRobot
_CMD_DPiROBOT__ADD_WAYPOINT                                    = 0x05    # robot_AddWaypoint
_CMD_DPiROBOT__ADD_WAYPOINT_ROBOT_COORDS                       = 0x06    # robot_AddWaypoint_RobotCoords
_CMD_DPiROBOT__CALCULATE_CURRENT_POSITION                      = 0x07    # robot_GetCurrentPosition
_CMD_DPiROBOT__CALCULATE_CURRENT_POSITION_ROBOT_COORDS         = 0x08    # robot_GetCurrentPosition_RobotCoords
_CMD_DPiROBOT__GET_CURRENT_POSITION                            = 0x09
_CMD_DPiROBOT__BUFFER_WAYPOINTS_BEFORE_STARTING                = 0X0a

_CMD_DPiROBOT__SET_ROBOT_TYPE                                  = 0x11    # robot_SetRobotType
_CMD_DPiROBOT__SET_ROBOT_DEFAULT_ACCELERATION                  = 0x12    # robot_SetRobotDefaultAcceleration
_CMD_DPiROBOT__SET_ROBOT_DEFAULT_JUNCTION_DEVIATION            = 0x13    # robot_SetRobotDefaultJunctionDeviation
_CMD_DPiROBOT__SET_ROBOT_MIN_MAX_X                             = 0x14    # robot_SetRobotMinMaxX
_CMD_DPiROBOT__SET_ROBOT_MIN_MAX_Y                             = 0x15    # robot_SetRobotMinMaxY
_CMD_DPiROBOT__SET_ROBOT_MIN_MAX_Z                             = 0x16    # robot_SetRobotMinMaxZ
_CMD_DPiROBOT__SET_HOMING_SPEED                                = 0x17    # robot_SetHomingSpeed
_CMD_DPiROBOT__SET_HOMING_METHOD                               = 0x18    # robot_SetHomingMethod
_CMD_DPiROBOT__SET_MAX_HOMING_DISTANCE_DEG_OR_MM               = 0x19    # robot_SetMaxHomingDistanceDegOrMM

_CMD_DPiROBOT__LIN_DELTA_SET_STEPS_PER_MM                      = 0x21    # robot_LinDelta_SetStepsPerMM
_CMD_DPiROBOT__LIN_DELTA_SET_ARM_LENGTH                        = 0x22    # robot_LinDelta_SetArmLength
_CMD_DPiROBOT__LIN_DELTA_SET_TOWER_AND_END_EFFECTOR_RADIUS     = 0x23    # robot_LinDelta_SetTowerAndEndEffectorRadius
_CMD_DPiROBOT__LIN_DELTA_SET_MAX_JOINT_POS                     = 0x24    # robot_LinDelta_SetMaxJointPos

_CMD_DPiROBOT__ROT_DELTA_SET_STEPS_PER_DEGREE                  = 0x31    # robot_RotDelta_SetStepsPerDegree
_CMD_DPiROBOT__ROT_DELTA_SET_ARM_LENGTH                        = 0x32    # robot_RotDelta_SetArmLengths
_CMD_DPiROBOT__ROT_DELTA_SET_BASE_AND_END_EFFECTOR_RADIUS      = 0x33    # robot_RotDelta_SetBaseAndEndEffectorRadius
_CMD_DPiROBOT__ROT_DELTA_SET_UPPER_ARM_HOME_ANGLE_DEGREES      = 0x34    # robot_RotDelta_SetUpperArmHomeAngleDegrees
_CMD_DPiROBOT__ROT_DELTA_SET_MIN_MAX_JOINT_ANGLES              = 0x35    # robot_RotDelta_SetMinMaxJointAngles

_CMD_DPiROBOT__MOTOR_DRIVER_SET_DRIVER_TYPE                    = 0x41    # motorDriver_SetDriverType
_CMD_DPiROBOT__MOTOR_DRIVER_SET_DRIVER_MICROSTEPPING           = 0x42    # motorDriver_SetDriverMicrostepping
_CMD_DPiROBOT__MOTOR_DRIVER_SET_REVERSE_DIRECTION_FLAG         = 0x43    # motorDriver_SetReverseStepDirectionFlag


#
# other constants used by this class
#
_DPiNETWORK_TIMEOUT_PERIOD_MS = 5
_DPiNETWORK_BASE_ADDRESS = 0x14


class DPiRobot:
    #
    # values for CMD_DPiROBOT__SET_ROBOT_TYPE
    #
    ROBOT_TYPE_ROTATIONAL_DELTA = 1
    ROBOT_TYPE_LINEAR_DELTA = 2
    ROBOT_TYPE_CARTESIAN = 3

    #
    # values for motorDriverType
    #
    MOTOR_DRIVER_TYPE_DRV8825 = 1
    MOTOR_DRIVER_TYPE_DM542T = 2

    #
    # values for CMD_DPiROBOT__SET_HOMING_METHOD
    #
    HOMING_METHOD_LIMIT_SWITCHES = 1
    HOMING_METHOD_NONE = 3

    #
    # values for CMD_DPiROBOT__GET_ROBOT_STATE
    #
    STATE_COMMUNICATION_FAILURE = 1             # getting the robot's state failed
    STATE_NOT_READY = 2                         # robot is not ready to accept commands from the host
    STATE_MOTORS_DISABLED = 3                   # motors are disabled
    STATE_NOT_HOMED = 4                         # motors enabled but robot not homed (either hasn't homed or homing failed)
    STATE_HOMING = 5                            # robot is running the homing procedure 
    STATE_STOPPED = 6                           # robot is stopped, waiting for waypoints from the hose
    STATE_PREPARING_TO_MOVE = 7                 # robot has received 1 or more waypoints but hasn't started moving yet
    STATE_MOVING = 8                            # moving to next waypoint
    STATE_WAYPOINT_BUFFER_FULL_FLAG = 0x80      # this bit set in status byte if waypoint buffer is full

    #
    # status values for CMD_DPiROBOT__GET_CURRENT_POSITION
    #
    GET_CURRENT_POSITION_NOT_READY = 1
    GET_CURRENT_POSITION_UNKNOWN = 2
    GET_CURRENT_POSITION_READY = 3

    #
    # attributes local to this class
    #
    _slaveAddress = _DPiNETWORK_BASE_ADDRESS
    _commErrorCount = 0


    #
    # constructor for the DPiRobot class
    #
    def __init__(self):
        pass


    # ---------------------------------------------------------------------------------
    #                                 Private functions 
    # ---------------------------------------------------------------------------------

    #
    # send a command to the DPiRobot, command's additional data must have already been "Pushed". 
    # After this function returns data from the device is retrieved by "Popping"
    #    Enter:  command = command byte
    #    Exit:   True returned on success, else False
    #
    def __sendCommand(self, command: int):
        (results, failedCount) = dpiNetwork.sendCommand(self._slaveAddress, command, _DPiNETWORK_TIMEOUT_PERIOD_MS)
        self._commErrorCount += failedCount;
        return results

    #
    # push a float with 0 digits right of the decimal, as an int32_t, to the transmit buffer 
    #  Enter:  float value 
    #
    def __pushFltPoint0ToInt32(self, value: float):
        dpiNetwork.pushInt32(int(round(value)))

    #
    # push a float with 3 digits right of the decimal, as an int32_t, to the transmit buffer
    #  Enter:  float value 
    #
    def __pushFltPoint3ToInt32(self, value: float):
        dpiNetwork.pushInt32(int(round(value * 1000.0)))

    #
    # push a float with 1 digit right of the decimal, as an int16_t, to the transmit buffer
    #
    def __pushFltPoint1ToInt16(self, value: float):
        dpiNetwork.pushInt16(int(round(value * 10.0)))

    #
    # pop an int16_t, to a float with 1 digits right of the decimal, from the receive buffer
    #
    def __popInt16ToFltPoint1(self):
        return float(dpiNetwork.popInt16()) / 10.0

 
    # ---------------------------------------------------------------------------------
    #                     Public setup and configuration functions 
    # ---------------------------------------------------------------------------------

    #
    # set the DPiRobot board number
    #    Enter:  boardNumber = DPiRobot board number (0 - 3)
    #
    def setBoardNumber(self, boardNumber: int):
        self._slaveAddress = _DPiNETWORK_BASE_ADDRESS + boardNumber
        

    #
    # ping the board
    #    Exit:   True returned on success, else False
    #
    def ping(self):
        return self.__sendCommand(_CMD_DPiROBOT__PING)


    #
    # initialize the board to its "power on" configuration
    #    Exit:   True returned on success, else False
    #
    def initialize(self):
        return self.__sendCommand(_CMD_DPiROBOT__INITIALIZE)


    #
    # disable the stepper motors
    #    Exit:   True returned on success, else False
    #
    def disableMotors(self):
        return self.__sendCommand(_CMD_DPiROBOT__DISABLE_ROBOT_MOTORS)


    #
    # get the count of communication errors
    #    Exit:   0 return if no errors, else count of errors returned
    #
    def getCommErrorCount(self):
        return self._commErrorCount


    # ---------------------------------------------------------------------------------
    #           Public functions for getting the robot's status and position
    # ---------------------------------------------------------------------------------

    #
    # get the robot's status
    #  Exit:   [0]: True returned on success, else False
    #          [1]: Robot status:
    #                  .STATE_NOT_READY = robot is not ready to accept commands from the host
    #                  .STATE_MOTORS_DISABLED = motors are disabled
    #                  .STATE_NOT_HOMED = motors enabled but robot not homed (either hasn't homed or homing failed)
    #                  .STATE_HOMING = robot is running the homing procedure 
    #                  .STATE_STOPPED = robot is stopped, waiting for waypoints from the hose
    #                  .STATE_PREPARING_TO_MOVE = robot has received 1 or more waypoints but hasn't started moving yet
    #                  .STATE_MOVING = moving to next waypoint
    #
    def getRobotStatus(self):
        if self.__sendCommand(_CMD_DPiROBOT__GET_ROBOT_STATE) != True:
            return False, self.STATE_COMMUNICATION_FAILURE

        return True, dpiNetwork.popUint8() & 0x7f


    #
    # get the robot's status along with waypoint buffer full flag
    #  Exit:   [0]: True returned on success, else False
    #          [1]: Robot status:
    #                  .STATE_COMMUNICATION_FAILURE = getting the robot's state failed
    #                  .STATE_NOT_READY = robot is not ready to accept commands from the host
    #                  .STATE_MOTORS_DISABLED = motors are disabled
    #                  .STATE_NOT_HOMED = motors enabled but robot not homed (either hasn't homed or homing failed)
    #                  .STATE_HOMING = robot is running the homing procedure 
    #                  .STATE_STOPPED = robot is stopped, waiting for waypoints from the hose
    #                  .STATE_PREPARING_TO_MOVE = robot has received 1 or more waypoints but hasn't started moving yet
    #                  .STATE_MOVING = moving to next waypoint
    #          [2]: True returned waypoint buffer is full
    #
    def getRobotStatusWithWaypointBufferFullFlg(self):
        if self.__sendCommand(_CMD_DPiROBOT__GET_ROBOT_STATE) != True:
            return False, self.STATE_COMMUNICATION_FAILURE, False

        status = dpiNetwork.popUint8()
        if status & 0x80 != 0:
            waypointBufferFulFlg = True
        else:
            waypointBufferFulFlg = False
            
        status = status & 0x7f
        
        return True, status, waypointBufferFulFlg


    #
    # get robot's current absolute position in the User's Cartesian Coordinates System, these values 
    # updates while in motion, values are signed with units in millimeters
    # Note: This command executes in ~3ms on a Pi4 (1.4ms with the C library)
    #  Exit:   [0]: True returned on success, false on communication error or the robot doesn't know its position
    #          [1]: X coordinate
    #          [2]: Y coordinate
    #          [3]: Z coordinate
    #
    def getCurrentPosition(self):
        #
        # start by requesting that the robot calculates its position
        #
        if self.__sendCommand(_CMD_DPiROBOT__CALCULATE_CURRENT_POSITION) != True:
            return False, 0, 0, 0

        #
        # get the position from the robot, looping until the data is ready
        #
        for _i in range(6):
            if self.__sendCommand(_CMD_DPiROBOT__GET_CURRENT_POSITION) != True:
                return False, 0, 0, 0

            X =  self.__popInt16ToFltPoint1()
            Y =  self.__popInt16ToFltPoint1()
            Z =  self.__popInt16ToFltPoint1()
            getPositionStatus = dpiNetwork.popInt8()

            if getPositionStatus == self.GET_CURRENT_POSITION_READY:
                return True, X, Y, Z

            if getPositionStatus == self.GET_CURRENT_POSITION_UNKNOWN:
                return False, 0, 0, 0

        return False, 0, 0, 0


    #
    # get robot's current absolute position in the Robot's Cartesian Coordinates System, these values 
    # updates while in motion, values are signed with units in millimeters
    # Note: This command executes in ~3ms on a Pi4 (1.4ms with the C library)
    #  Exit:   [0]: True returned on success, false on communication error or the robot doesn't know its position
    #          [1]: X coordinate
    #          [2]: Y coordinate
    #          [3]: Z coordinate
    #
    def getCurrentPosition_RobotCoords(self):
        #
        # start by requesting that the robot calculates its position
        #
        if self.__sendCommand(_CMD_DPiROBOT__CALCULATE_CURRENT_POSITION_ROBOT_COORDS) != True:
            return False, 0, 0, 0

        #
        # get the position from the robot, looping until the data is ready
        #
        for _i in range(6):
            if self.__sendCommand(_CMD_DPiROBOT__GET_CURRENT_POSITION) != True:
                return False, 0, 0, 0

            X =  self.__popInt16ToFltPoint1()
            Y =  self.__popInt16ToFltPoint1()
            Z =  self.__popInt16ToFltPoint1()
            getPositionStatus = dpiNetwork.popInt8()

            if getPositionStatus == self.GET_CURRENT_POSITION_READY:
                return True, X, Y, Z

            if getPositionStatus == self.GET_CURRENT_POSITION_UNKNOWN:
                return False, 0, 0, 0

        return False, 0, 0, 0
 

    # ---------------------------------------------------------------------------------
    #                       Public functions that control the robot
    # ---------------------------------------------------------------------------------

    #
    # add a waypoint (using the User's Cartesian Coordinate System). This will return 
    # right away unless the waypoint buffer is full (in which case it will busy-wait).
    # Note: this command executes in ~4ms on a Pi4 (2.8ms for the C version of this library)
    #    Enter:  X, Y, Z = position to move to, values are signed with units in 
    #               millimeters  (−3276.8mm to 3276.7mm)
    #            speed = speed to travel in mm/sec
    #    Exit:   True returned on success, else False
    #
    def addWaypoint(self, X: float, Y: float, Z: float, speed: float):
        #
        # loop if waypoint buffer is full
        #
        while True:
            #
            # get the robot's status including if the waypoint buffer is full
            #
            results, _status, waypointBufferFullFlg = self.getRobotStatusWithWaypointBufferFullFlg()
            if results != True:
                return False

            #
            # stop waiting if the waypoint buffer is not full
            #
            if waypointBufferFullFlg == False:
                break

        #
        # there is room in the buffer, send the waypoint now
        #
        self.__pushFltPoint1ToInt16(X)
        self.__pushFltPoint1ToInt16(Y)
        self.__pushFltPoint1ToInt16(Z)
        self.__pushFltPoint1ToInt16(speed)
        return self.__sendCommand(_CMD_DPiROBOT__ADD_WAYPOINT)


    #
    # add a waypoint (using the Robot's Cartesian Coordinate System). This will return 
    # right away unless the waypoint buffer is full (in which case it will busy-wait).
    # Note: this command executes in ~4ms on a Pi4 
    #    Enter:  X, Y, Z = position to move to, values are signed with units in 
    #               millimeters  (−3276.8mm to 3276.7mm)
    #            speed = speed to travel in mm/sec
    #    Exit:   True returned on success, else False
    #
    def addWaypoint_RobotCoords(self, X: float, Y: float, Z: float, speed: float):
        #
        # loop if waypoint buffer is full
        #
        while True:
            #
            # get the robot's status including if the waypoint buffer is full
            #
            results, _status, waypointBufferFullFlg = self.getRobotStatusWithWaypointBufferFullFlg()
            if results != True:
                return False

            #
            # stop waiting if the waypoint buffer is not full
            #
            if waypointBufferFullFlg == False:
                break

        #
        # there is room in the buffer, send the waypoint now
        #
        self.__pushFltPoint1ToInt16(X)
        self.__pushFltPoint1ToInt16(Y)
        self.__pushFltPoint1ToInt16(Z)
        self.__pushFltPoint1ToInt16(speed)
        return self.__sendCommand(_CMD_DPiROBOT__ADD_WAYPOINT_ROBOT_COORDS)


    #
    # in some situations buffering waypoints can result in a more "fluid" motion,
    # preventing the robot from briefly stopping between each point as it moves  
    # along a continous path; to use: 1) while the robot is stopped, call this
    # function to enable buffering  2) send the robot all the waypoints along
    # the path  3)after sending the final waypoint immediately call this function
    # again to stop buffering.
    #
    def bufferWaypointsBeforeStartingToMove(self, bufferFlg: bool):
        dpiNetwork.pushUint8(bufferFlg)
        return self.__sendCommand(_CMD_DPiROBOT__BUFFER_WAYPOINTS_BEFORE_STARTING)


    #
    # wait while the robot is moving, return when it has stopped
    #    Exit:   True returned on success, else False
    #
    def waitWhileRobotIsMoving(self):
        self.bufferWaypointsBeforeStartingToMove(False)      # stop buffering waypoints if that was enabled

        while True:
            #
            # get the robot's status
            #
            results, robotStatus = self.getRobotStatus()
            if results != True:
                return False

            #
            # return if status indicates the robot has stopped
            #
            if not ((robotStatus == self.STATE_MOVING) or (robotStatus == self.STATE_PREPARING_TO_MOVE) or
                    (robotStatus == self.STATE_HOMING)):
                return True


    #
    # home the robot, if motors are disabled then enable them first
    #    Enter:  alwaysHomeFlg = True to always run homing procedure
    #                          = False to only run homing procedure if needed
    #    Exit:   True returned on success, else False
    #
    def homeRobot(self, alwaysHomeFlg: bool):
        #
        # verify communication with the robot is working
        #
        results, robotStatus = self.getRobotStatus()
        if (results != True) or (robotStatus == self.STATE_NOT_READY):
            return False

        #
        # start the robot homing
        #
        dpiNetwork.pushUint8(alwaysHomeFlg)
        if self.__sendCommand(_CMD_DPiROBOT__HOME_ROBOT) != True:
            return False

        #
        # check the robot's status until homing is complete
        #
        while True:
            results, robotStatus = self.getRobotStatus()
            if results != True:
                return False

            if robotStatus != self.STATE_HOMING:
                break

        #
        # check if homing failed
        #
        if robotStatus != self.STATE_STOPPED:
            return False

        #
        # homing successful
        #
        return True





#      ******************************************************************
#      *                                                                *
#      *                        DPiStepper Libary                       *
#      *                                                                *
#      *            Stan Reifel                     8/21/2022           *
#      *                                                                *
#      ******************************************************************

#
# This library is used to drive the DPiStepper board which can control up to 3
# stepper motors.  The motors are accelerated and decelerated as they travel 
# to the final position.  This driver supports changing the target position, 
# speed or rate of acceleration while moving.
#
# Running just one stepper motor, DPiStepper can generate a maximum of about 35K  
# steps per second.  Running one motor with 8X microstepping, the maximum rotation  
# rate is about 22 RPS or 1300 RPM.  Running multiple motors at the same time will 
# reduce the maximum speed.  For example running two motors will reduce the step 
# rate by a bit more than half.
# 
#
# Usage:
#    Near the top of the program, add:
#        from DPiStepper import DPiStepper
#
#    Declare a global object for the DPiStepper board functions as: 
#        dpiStepper = DPiStepper()
#
#    Jumpers on the DPiStepper board allow you to have up to 4 boards on the same
#    RS485 network. Call this function to tell the library which DPi stepper board 
#    you want to talk to.  In this example board 0 is selected:
#        dpiStepper.setBoardNumber(0)
#
#    Each DPiStepper board can control up to 3 stepper motors.  This example will setup
#    and move "Motor 0":
#        #
#        # enable the stepper motors, when disabled the motors are turned off and spin freely
#        #
#        dpiStepper.enableMotors(True)
#
#        #
#        # set the microstepping to 1, 2, 4, 8, 16 or 32. 8 results in 1600 steps per
#        # revolution of the motor's shaft
#        #
#        dpiStepper.setMicrostepping(8)
#
#        #
#        # set the speed in steps/second and acceleration in steps/second/second
#        #
#        stepperNum = 0
#        dpiStepper.setSpeedInStepsPerSecond(stepperNum, 1600)
#        dpiStepper.setAccelerationInStepsPerSecondPerSecond(stepperNum, 1600)
#
#        #
#        # move 1600 steps in the backward direction, this function will return after the
#        # motor stops because "wait_to_finish_moving_flg" is set to True
#        #
#        wait_to_finish_moving_flg = True
#        dpiStepper.moveToRelativePositionInSteps(stepper_num, -1600, wait_to_finish_moving_flg)
#
#        #
#        # move to an absolute position of 1600 steps
#        #
#        dpiStepper.moveToAbsolutePositionInSteps(stepperNum, 1600, wait_to_finish_moving_flg)
#
#
#    There are 3 different units that can be used to control the stepper motors: (steps, 
#    millimeters or revolutions).  Choose the one that's most intuitive for your 
#    application.  This example will move "Motor 1" using units of Revolutions:
#        #
#        # set the number of steps per revolutions, 200 with no microstepping, 
#        # 800 with 4x microstepping
#        #
#        stepperNum = 1
#        dpiStepper.setStepsPerRevolution(stepperNum, 800)
#
#        #
#        # set the speed in rotations/second and acceleration in 
#        # rotations/second/second
#        #
#        dpiStepper.setSpeedInRevolutionsPerSecond(stepperNum, 1.0)
#        dpiStepper.setAccelerationInRevolutionsPerSecondPerSecond(stepperNum, 1.0)
#
#        #
#        # move backward 1.5 revolutions, this function will return after the
#        # motor stops because "wait_to_finish_moving_flg" is set to True
#        #
#        wait_to_finish_moving_flg = True
#        dpiStepper.moveToRelativePositionInRevolutions(stepperNum, -1.5, wait_to_finish_moving_flg)
#
#        #
#        # move to an absolute position of 3.75 revolutions
#        #
#        dpiStepper.moveToAbsolutePositionInRevolutions(stepperNum, 3.75, wait_to_finish_moving_flg)
#
#
#    For linear mechanisms, controlling the stepper motors with units in millimeters
#    often makes the most sense.  This example is driving "Motor 2" in millimeters:
#        #
#        # set the number of steps per millimeter
#        #
#        stepperNum = 2
#        dpiStepper.setStepsPerMillimeter(stepperNum, 25.0)
#
#        #
#        # set the speed in millimeters/second and acceleration in 
#        # millimeters/second/second
#        #
#        dpiStepper.setSpeedInMillimetersPerSecond(stepperNum, 20.0)
#        dpiStepper.setAccelerationInMillimetersPerSecondPerSecond(stepperNum, 20.0)
#
#        #
#        # move backward 15.5 millimeters
#        #
#        wait_to_finish_moving_flg = True
#        dpiStepper.moveToRelativePositionInMillimeters(stepperNum, -15.5, wait_to_finish_moving_flg)
#
#        #
#        # move to an absolute position of 125 millimeters
#        #
#        dpiStepper.moveToAbsolutePositionInMillimeters(stepperNum, 125, wait_to_finish_moving_flg)
#


from array import *
from time import sleep
from DPiNetwork import DPiNetwork
dpiNetwork = DPiNetwork()

#
# DPiNetwork DPiStepper commands  
#
_CMD_DPiSTEPPER__PING                                  = 0x00
_CMD_DPiSTEPPER__INITIALIZE                            = 0x01
_CMD_DPiSTEPPER__ENABLE_MOTORS                         = 0x02
_CMD_DPiSTEPPER__DISABLE_MOTORS                        = 0x03
_CMD_DPiSTEPPER__SET_MICROSTEPPING                     = 0x04
_CMD_DPiSTEPPER__ALL_MOTORS_STOPPED                    = 0x05
_CMD_DPiSTEPPER__SET_CURRENT_P0SITION_IN_STEPS         = 0x10
_CMD_DPiSTEPPER__GET_CURRENT_POSITION_IN_STEPS         = 0x20
_CMD_DPiSTEPPER__SET_SPEED_IN_STEPS_PER_SEC            = 0x30
_CMD_DPiSTEPPER__SET_ACCEL_IN_STEPS_PSPS               = 0x40
_CMD_DPiSTEPPER__MOVE_TO_ABS_POSITION_IN_STEPS         = 0x50
_CMD_DPiSTEPPER__MOVE_TO_REL_POSITION_IN_STEPS         = 0x60
_CMD_DPiSTEPPER__DECELERATE_TO_A_STOP                  = 0x70
_CMD_DPiSTEPPER__EMERGENCY_STOP                        = 0x80
_CMD_DPiSTEPPER__GET_CURRENT_VELOCITY_IN_STEPS_PER_SEC = 0x90
_CMD_DPiSTEPPER__MOTION_COMPLETE                       = 0xa0
_CMD_DPiSTEPPER__GET_STEPPER_STATUS                    = 0xb0


#
# bits returned by CMD_DPiSTEPPER__GET_STEPPER_STATUS
#
_STEPPER_STATUS_MOTOR_STOPPED  = 0x01
_STEPPER_STATUS_MOTORS_ENABLED = 0x02
_STEPPER_STATUS_AT_HOME_SWITCH = 0x04


#
# other constants used by this class
#
_NUMBER_OF_DPi_STEPPER_DRIVERS = 3
_DPiNETWORK_TIMEOUT_PERIOD_MS = 6
_DPiNETWORK_BASE_ADDRESS = 0x20



class DPiStepper:
    #
    # audio wave shapes
    #
    SINE_WAVE       = 0
    TRIANGLE_WAVE   = 1
    SQUARE_WAVE     = 2

    #
    # attributes local to this class
    #
    _slaveAddress = _DPiNETWORK_BASE_ADDRESS
    _commErrorCount = 0
    _stepsPerMillimeter = array('d', [200.0, 200.0, 200.0])
    _stepsPerRevolution = array('d', [1600.0, 1600.0, 1600.0])


    #
    # constructor for the DPiStepper class
    #
    def __init__(self):
        pass
       

    # ---------------------------------------------------------------------------------
    #                                 Private functions 
    # ---------------------------------------------------------------------------------

    #
    # send a command to the DPiStepper, command's additional data must have already been "Pushed". 
    # After this function returns data from the device is retrieved by "Popping"
    #    Enter:  command = command byte
    #    Exit:   True returned on success, else False
    #
    def __sendCommand(self, command: int):
        (results, failedCount) = dpiNetwork.sendCommand(self._slaveAddress, command, _DPiNETWORK_TIMEOUT_PERIOD_MS)
        self._commErrorCount += failedCount;
        return results


    #
    # push a float with 2 digits right of the decimal, as an int32_t, to the transmit buffer
    #  Enter:  float value 
    #
    def __pushFltPoint2ToInt32(self, value: float):
        dpiNetwork.pushInt32(int(round(value * 100.0)))


    #
    # pop an int32_t, as a float with 2 digits right of the decimal, from the receive buffer
    #
    def __popInt32ToFltPoint2(self):
        return float(dpiNetwork.popInt32()) / 100.0


    # ---------------------------------------------------------------------------------
    #                     Public setup and configuration functions 
    # ---------------------------------------------------------------------------------

    #
    # set the DPiStepper board number
    #    Enter:  boardNumber = DPiStepper board number (0 - 3)
    #
    def setBoardNumber(self, boardNumber: int):
        self._slaveAddress = _DPiNETWORK_BASE_ADDRESS + boardNumber


    #
    # ping the board
    #    Exit:   True returned on success, else False
    #
    def ping(self):
        return self.__sendCommand(_CMD_DPiSTEPPER__PING)


    #
    # initialize the board to its "power on" configuration
    #    Exit:   True returned on success, else False
    #
    def initialize(self):
        for i in range(_NUMBER_OF_DPi_STEPPER_DRIVERS):
            self._stepsPerMillimeter[i] = 200
            self._stepsPerRevolution[i] = 1600
        return self.__sendCommand(_CMD_DPiSTEPPER__INITIALIZE)


    #
    # enable/disable the stepper motors
    #  Exit:   True returned on success, else False
    #
    def enableMotors(self, enableFlg: bool):
        if enableFlg:
            results = self.__sendCommand(_CMD_DPiSTEPPER__ENABLE_MOTORS)
            sleep(.1);
            return results
        else:
            return self.__sendCommand(_CMD_DPiSTEPPER__DISABLE_MOTORS)


    #
    # set microstepping 
    #    Enter:  microstepping = 1, 2, 4, 8, 16, 32
    #    Exit:   True returned on success, else False
    #
    def setMicrostepping(self, microstepping: int):
        if not ((microstepping == 1) or (microstepping == 2) or (microstepping == 4) or
                (microstepping == 8) or (microstepping == 16) or (microstepping == 32)):
            return False

        dpiNetwork.pushUint8(microstepping)
        return self.__sendCommand(_CMD_DPiSTEPPER__SET_MICROSTEPPING)


    # ---------------------------------------------------------------------------------
    #                             Public status functions 
    # ---------------------------------------------------------------------------------

    #
    # check if all 3 stepper motors stopped
    #  Exit:   True returned if motion complete, else False
    #
    def getAllMotorsStopped(self):
        if self.__sendCommand(_CMD_DPiSTEPPER__ALL_MOTORS_STOPPED) != True:
            return False

        allStoppedFlg = dpiNetwork.popUint8()

        if allStoppedFlg:
            return True
        else:
            return False


    #
    # wait until motor stops
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   True returned on success, else False
    #
    def waitUntilMotorStops(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, False

        while True:
            results, stoppedFlg = self.getMotionComplete(stepperNum)
            if results != True:
                return False
            if stoppedFlg == True:
                return True
            sleep(0.02)


    #
    # get motion complete flag, indicating that the stepper has stopped
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: True returned on motion complete, else False
    #
    def getMotionComplete(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, False

        if self.__sendCommand(_CMD_DPiSTEPPER__MOTION_COMPLETE + stepperNum) != True:
            return False, False

        if dpiNetwork.popUint8() == False:
            return True, False
        else:
            return True, True



    #
    # get stepper status
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: True returned if motor is stopped
    #          [2]: True returned if motors are enabled
    #          [3]: True returned if the "Homing" switch indicates "At home"
    #
    def getStepperStatus(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, False, False, False

        if self.__sendCommand(_CMD_DPiSTEPPER__GET_STEPPER_STATUS + stepperNum) != True:
            return False, False, False, False

        results = dpiNetwork.popUint8()
        if results & _STEPPER_STATUS_MOTOR_STOPPED != 0:
            stoppedFlg = True
        else: 
            stoppedFlg = False

        if results & _STEPPER_STATUS_MOTORS_ENABLED != 0:
            motorsEnabledFlg = True
        else:
            motorsEnabledFlg = False

        if results & _STEPPER_STATUS_AT_HOME_SWITCH:
            homeAtHomeSwitchFlg = True
        else:
            homeAtHomeSwitchFlg = False

        return True, stoppedFlg, motorsEnabledFlg, homeAtHomeSwitchFlg


 
    #
    # get the count of communication errors
    #    Exit:   0 return if no errors, else count of errors returned
    #
    def getCommErrorCount(self):
        return self._commErrorCount


    # ---------------------------------------------------------------------------------
    #                         Public functions with units in Steps
    # ---------------------------------------------------------------------------------

    #
    # set the maximum speed with units in steps/second, this is the maximum speed reached  
    # while accelerating
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          speedInStepsPerSecond = speed to accelerate up to, units in steps/second
    #  Exit:   True returned on success, else False
    #
    def setSpeedInStepsPerSecond(self, stepperNum: int, speedInStepsPerSecond: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        if (speedInStepsPerSecond < 0.05) or (speedInStepsPerSecond > 30000.0):
            return False

        self.__pushFltPoint2ToInt32(speedInStepsPerSecond)
        return self.__sendCommand(_CMD_DPiSTEPPER__SET_SPEED_IN_STEPS_PER_SEC + stepperNum)


    #
    # set the rate of acceleration with units in steps/second/second
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          accelerationInStepsPerSecondPerSecond = rate of acceleration, units in 
    #            steps/second/second
    #  Exit:   True returned on success, else False
    #
    def setAccelerationInStepsPerSecondPerSecond(self, stepperNum: int, accelerationInStepsPerSecondPerSecond: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        if (accelerationInStepsPerSecondPerSecond < 0.05) or (accelerationInStepsPerSecondPerSecond > 30000.0):
            return False

        self.__pushFltPoint2ToInt32(accelerationInStepsPerSecondPerSecond)
        return self.__sendCommand(_CMD_DPiSTEPPER__SET_ACCEL_IN_STEPS_PSPS + stepperNum)



    #
    # set the current position of the motor in Steps, this does not move the motor
    # Note: This function should only be called when the motor is stopped
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          currentPositionInSteps = the new position of the motor in steps
    #  Exit:   True returned on success, else False
    #
    def setCurrentPositionInSteps(self, stepperNum: int, currentPositionInSteps: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        dpiNetwork.pushInt32(currentPositionInSteps)
        return self.__sendCommand(_CMD_DPiSTEPPER__SET_CURRENT_P0SITION_IN_STEPS + stepperNum)



    #
    # get the current position of the motor in Steps, this functions is updated
    # while the motor moves
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          currentPositionInSteps -> storage to return current position in
    #  Exit:   [0]: True returned on success, else False
    #          [1]: a signed motor position in steps returned
    #
    def getCurrentPositionInSteps(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, 0

        if self.__sendCommand(_CMD_DPiSTEPPER__GET_CURRENT_POSITION_IN_STEPS + stepperNum) != True:
            return False, 0

        currentPositionInSteps = dpiNetwork.popInt32()
        return True, currentPositionInSteps


    #
    # move to an absolute position, units are in Steps
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          absolutePositionToMoveToInSteps = signed absolute position to move to  
    #            in units of steps
    #          waitToFinishFlg = True to wait before returning until motor has stops
    #  Exit:   True returned on success, else False
    #
    def moveToAbsolutePositionInSteps(self, stepperNum: int, absolutePositionToMoveToInSteps: int, waitToFinishFlg: bool):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        dpiNetwork.pushInt32(absolutePositionToMoveToInSteps)
        results = self.__sendCommand(_CMD_DPiSTEPPER__MOVE_TO_ABS_POSITION_IN_STEPS + stepperNum)
        if results != True:
            return False
        
        if waitToFinishFlg:
            return self.waitUntilMotorStops(stepperNum)
        else:
            return True


    #
    # move a relative distance from the current position, distance to move is in Steps
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          distanceToMoveInSteps = signed distance to move relative to the current  
    #            position in steps
    #          waitToFinishFlg = True to wait before returning until motor has stops
    #  Exit:   True returned on success, else False
    #
    def moveToRelativePositionInSteps(self, stepperNum: int, distanceToMoveInSteps: int, waitToFinishFlg: bool):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        dpiNetwork.pushInt32(distanceToMoveInSteps)
        results = self.__sendCommand(_CMD_DPiSTEPPER__MOVE_TO_REL_POSITION_IN_STEPS + stepperNum)
        if results != True:
            return False
        
        if waitToFinishFlg:
            return self.waitUntilMotorStops(stepperNum)
        else:
            return True



    #
    # Get the current velocity of the motor in steps/second.  This functions is 
    # updated while it accelerates up and down in speed.  This is not the desired  
    # speed, but the speed the motor should be moving at the time the function is  
    # called.  This is a signed value and is negative when the motor is moving 
    # backwards.  Note: This speed will be incorrect if the desired velocity is set 
    # faster than this library can generate steps, or if the load on the motor is too 
    # great for the amount of torque that it can generate.
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: current velocity in steps/second, signed
    #
    def getCurrentVelocityInStepsPerSecond(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, 0
    
        results = self.__sendCommand(_CMD_DPiSTEPPER__GET_CURRENT_VELOCITY_IN_STEPS_PER_SEC + stepperNum)
        if results != True:
            return False, 0

        currentVelocity = self.__popInt32ToFltPoint2()
        return True, currentVelocity


    #
    # decelerate the stepper motor until it stops
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   True returned on success, else False
    #
    def decelerateToAStop(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.__sendCommand(_CMD_DPiSTEPPER__DECELERATE_TO_A_STOP + stepperNum)


    #
    # emergency stop - just stop stepping the motor, expect steps will be lost so motor should be re-homed
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   True returned on success, else False
    #
    def emergencyStop(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False
    
        return self.__sendCommand(_CMD_DPiSTEPPER__EMERGENCY_STOP + stepperNum)


    #
    # home the motor by moving until the homing sensor is activated, then set the 
    # position to zero with units in steps
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          directionTowardHome = 1 to move in a positive direction, -1 to move in 
    #            a negative directions 
    #          speedInStepsPerSecond = speed to accelerate up to while moving toward 
    #            home, units in steps/second
    #          maxDistanceToMoveInSteps = unsigned maximum distance to move toward 
    #            home before giving up
    #  Exit:   True returned if successful, else False
    #
    def moveToHomeInSteps(self, stepperNum: int, directionTowardHome: int,
                          speedInStepsPerSecond: float, maxDistanceToMoveInSteps: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        if not ((directionTowardHome == 1) or (directionTowardHome == -1)):
            return False

        #
        # enable the motors
        #
        if self.enableMotors(True) != True:
            return False

        #
        # set the homing speed
        #
        if self.setSpeedInStepsPerSecond(stepperNum, speedInStepsPerSecond) != True:
            return False
        if self.setAccelerationInStepsPerSecondPerSecond(stepperNum, speedInStepsPerSecond) != True:
            return False
         
        #
        # if not at the home switch alread, move toward it
        #
        results, stoppedFlg, __, homeAtHomeSwitchFlg = self.getStepperStatus(stepperNum)
        if results != True:
            return False

        if homeAtHomeSwitchFlg != True:
            #
            # move toward the home switch
            #
            if self.moveToRelativePositionInSteps(stepperNum, maxDistanceToMoveInSteps * directionTowardHome, False) != True:
                return False

            #
            # while motor is moving, check if limit switch is hit, or if travels the "max" distance
            #
            while True:
                results, stoppedFlg, __, homeAtHomeSwitchFlg = self.getStepperStatus(stepperNum)
                if results != True:
                    return False
                if stoppedFlg:
                    return False
                if homeAtHomeSwitchFlg:
                    break

            self.emergencyStop(stepperNum)
            sleep(.1)


        #
        # the switch has been detected, now move away from the switch
        #
        if self.moveToRelativePositionInSteps(stepperNum, -maxDistanceToMoveInSteps * directionTowardHome, False) != True:
            return False

        #
        # while motor is moving, check if limit switch is released, or if travels the "max" distance
        #
        while True:
            results, stoppedFlg, __, homeAtHomeSwitchFlg = self.getStepperStatus(stepperNum)
            if results != True:
                return False
            if stoppedFlg:
                return False
            if homeAtHomeSwitchFlg != True:
                break
    
        self.emergencyStop(stepperNum)
        sleep(.1)


        #
        # have now moved off the switch, move toward it again but slower
        #
        if self.setSpeedInStepsPerSecond(stepperNum, speedInStepsPerSecond/8) != True:
            return False
        if self.moveToRelativePositionInSteps(stepperNum, maxDistanceToMoveInSteps * directionTowardHome, False) != True:
            return False

        while True:
            results, stoppedFlg, __, homeAtHomeSwitchFlg = self.getStepperStatus(stepperNum)
            if results != True:
                return False
            if stoppedFlg:
                return False
            if homeAtHomeSwitchFlg:
                break

        self.emergencyStop(stepperNum)
        sleep(.1)
        

        #
        # successfully homed, set the current position to 0
        #
        if self.setCurrentPositionInSteps(stepperNum, 0) != True:
            return False

        return True


    # ---------------------------------------------------------------------------------
    #                     Public functions with units in Millimeters
    # ---------------------------------------------------------------------------------
    
    #
    # set the number of steps the per millimeters of motion
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          stepsPerMillimeter = number of motor steps per millimeter of travel
    #
    def setStepsPerMillimeter(self, stepperNum: int, motorStepsPerMillimeter: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return
        self._stepsPerMillimeter[stepperNum] = motorStepsPerMillimeter


    #
    # set the current position in millimeters, this does not move the motor
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          newPositionInMillimeters = value in MM to set current position
    #  Exit:   True returned on success, else False
    #
    def setCurrentPositionInMillimeters(self, stepperNum: int, newPositionInMillimeters: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        positionInSteps = int(round(newPositionInMillimeters * self._stepsPerMillimeter[stepperNum]))
        return self.setCurrentPositionInSteps(stepperNum, positionInSteps)


    #
    # get the current position in millimeters, this functions is updated
    # while the motor moves
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: a signed position in millimeters
    #
    def getCurrentPositionInMillimeters(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, 0.0

        results, positionInSteps = self.getCurrentPositionInSteps(stepperNum)
        if results != True:
            return False

        return True, float(positionInSteps) / self._stepsPerMillimeter[stepperNum]
   

    #
    # set the maximum speed with units in millimeters/second, this is the maximum speed  
    # reached while accelerating
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          speedInMillimetersPerSecond = speed to accelerate up to, units in 
    #            millimeters/second
    #  Exit:   True returned on success, else False
    #
    def setSpeedInMillimetersPerSecond(self, stepperNum: int, speedInMillimetersPerSecond: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.setSpeedInStepsPerSecond(stepperNum, int(speedInMillimetersPerSecond * self._stepsPerMillimeter[stepperNum]))


    #
    # set the rate of acceleration, units in millimeters/second/second
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Enter:  accelerationInMillimetersPerSecondPerSecond = rate of acceleration,  
    #          units in millimeters/second/second
    #  Exit:   True returned on success, else False
    #
    def setAccelerationInMillimetersPerSecondPerSecond(self, stepperNum: int, accelerationInMillimetersPerSecondPerSecond: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.setAccelerationInStepsPerSecondPerSecond(stepperNum,
                    int(accelerationInMillimetersPerSecondPerSecond * self._stepsPerMillimeter[stepperNum]))


    #
    # move to an absolute position, units are in millimeters
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          absolutePositionToMoveToInMillimeters = signed absolute position to move
    #          waitToFinishFlg = True to wait before returning until motor has stops
    #  Exit:   True returned on success, else False
    #
    def moveToAbsolutePositionInMillimeters(self, stepperNum: int, absolutePositionToMoveToInMillimeters: float, waitToFinishFlg: bool):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.moveToAbsolutePositionInSteps(stepperNum, 
                  int(round(absolutePositionToMoveToInMillimeters * self._stepsPerMillimeter[stepperNum])), waitToFinishFlg)

 
    #
    # move a relative distance from the current position, distance to move is in millimeters
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          distanceToMoveInMillimeters = signed distance to move relative
    #          waitToFinishFlg = True to wait before returning until motor has stops
    #  Exit:   True returned on success, else False
    #
    def moveToRelativePositionInMillimeters(self, stepperNum: int, distanceToMoveInMillimeters: float, waitToFinishFlg: bool):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.moveToRelativePositionInSteps(stepperNum, 
                  int(round(distanceToMoveInMillimeters * self._stepsPerMillimeter[stepperNum])), waitToFinishFlg)


    #
    # Get the current velocity of the motor in Millimeters/second.  This functions 
    # is updated while it accelerates up and down in speed.  This is not the desired  
    # speed, but the speed the motor should be moving at the time the function is  
    # called.  This is a signed value and is negative when the motor is moving 
    # backwards.  Note: This speed will be incorrect if the desired velocity is set 
    # faster than this library can generate steps, or if the load on the motor is too 
    # great for the amount of torque that it can generate.
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: current velocity in millimeters/second, signed
    #
    def getCurrentVelocityInMillimetersPerSecond(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, 0
    
        results, velocityInStepsPerSec = self.getCurrentVelocityInStepsPerSecond(stepperNum)
        if results != True:
            return False, 0

        return True, velocityInStepsPerSec / self._stepsPerMillimeter[stepperNum]


    #
    # home the motor by moving until the homing sensor is activated, then set the 
    # position to zero, with units in Millimeters
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          directionTowardHome = 1 to move in a positive direction, -1 to move in 
    #            a negative directions 
    #          speedInMillimetersPerSecond = speed to accelerate up to while moving 
    #            toward home, units in millimeters/second
    #          maxDistanceToMoveInMillimeters = unsigned maximum distance to move 
    #            toward home before giving up
    #  Exit:   True returned on success, else False
    #
    def moveToHomeInMillimeters(self, stepperNum: int, directionTowardHome: int, 
       speedInMillimetersPerSecond: float, maxDistanceToMoveInMillimeters: float):
        
        return self.moveToHomeInSteps(stepperNum, directionTowardHome,
                                       int(speedInMillimetersPerSecond * self._stepsPerMillimeter[stepperNum]),
                                       int (maxDistanceToMoveInMillimeters * self._stepsPerMillimeter[stepperNum]))


    # ---------------------------------------------------------------------------------
    #                     Public functions with units in Revolutions
    # ---------------------------------------------------------------------------------

    #
    # set the number of steps per Revolutions
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          motorStepPerRevolution = number of motor steps per revolution of travel
    #
    def setStepsPerRevolution(self, stepperNum: int, motorStepPerRevolution: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return
        self._stepsPerRevolution[stepperNum] = motorStepPerRevolution


    #
    # set the current position in revolutions, this does not move the motor
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          newPositionInRevolutions = value in Revolutions to set current position
    #  Exit:   True returned on success, else False
    #
    def setCurrentPositionInRevolutions(self, stepperNum: int, newPositionInRevolutions: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        positionInSteps = int(round(newPositionInRevolutions * self._stepsPerRevolution[stepperNum]))
        return self.setCurrentPositionInSteps(stepperNum, positionInSteps)


    #
    # get the current position in revolutions, this functions is updated
    # while the motor moves
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: a signed position in revolutions
    #
    def getCurrentPositionInRevolutions(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, 0.0

        results, positionInSteps = self.getCurrentPositionInSteps(stepperNum)
        if results != True:
            return False

        return True, float(positionInSteps) / self._stepsPerRevolution[stepperNum]
   

    #
    # set the maximum speed with units in revolutions/second, this is the maximum speed  
    # reached while accelerating
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          speedInRevolutionsPerSecond = speed to accelerate up to, units in 
    #            revolutions/second
    #  Exit:   True returned on success, else False
    #
    def setSpeedInRevolutionsPerSecond(self, stepperNum: int, speedInRevolutionsPerSecond: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.setSpeedInStepsPerSecond(stepperNum, int(speedInRevolutionsPerSecond * self._stepsPerRevolution[stepperNum]))


    #
    # set the rate of acceleration, units in revolutions/second/second
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Enter:  accelerationInRevolutionsPerSecondPerSecond = rate of acceleration,  
    #          units in revolutions/second/second
    #  Exit:   True returned on success, else False
    #
    def setAccelerationInRevolutionsPerSecondPerSecond(self, stepperNum: int, accelerationInRevolutionsPerSecondPerSecond: float):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.setAccelerationInStepsPerSecondPerSecond(stepperNum,
                    int(accelerationInRevolutionsPerSecondPerSecond * self._stepsPerRevolution[stepperNum]))


    #
    # move to an absolute position, units are in revolutions
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          absolutePositionToMoveToInRevolutions = signed absolute position to move
    #          waitToFinishFlg = True to wait before returning until motor has stops
    #  Exit:   True returned on success, else False
    #
    def moveToAbsolutePositionInRevolutions(self, stepperNum: int, absolutePositionToMoveToInRevolutions: float, waitToFinishFlg: bool):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.moveToAbsolutePositionInSteps(stepperNum, 
                  int(round(absolutePositionToMoveToInRevolutions * self._stepsPerRevolution[stepperNum])), waitToFinishFlg)

 
    #
    # move a relative distance from the current position, distance to move is in revolutions
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          distanceToMoveInRevolutions = signed distance to move relative
    #          waitToFinishFlg = True to wait before returning until motor has stops
    #  Exit:   True returned on success, else False
    #
    def moveToRelativePositionInRevolutions(self, stepperNum: int, distanceToMoveInRevolutions: float, waitToFinishFlg: bool):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False

        return self.moveToRelativePositionInSteps(stepperNum, 
                  int(round(distanceToMoveInRevolutions * self._stepsPerRevolution[stepperNum])), waitToFinishFlg)


    #
    # Get the current velocity of the motor in Revolutions/second.  This functions 
    # is updated while it accelerates up and down in speed.  This is not the desired  
    # speed, but the speed the motor should be moving at the time the function is  
    # called.  This is a signed value and is negative when the motor is moving 
    # backwards.  Note: This speed will be incorrect if the desired velocity is set 
    # faster than this library can generate steps, or if the load on the motor is too 
    # great for the amount of torque that it can generate.
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #  Exit:   [0]: True returned on success, else False
    #          [1]: current velocity in revolutions/second, signed
    #
    def getCurrentVelocityInRevolutionsPerSecond(self, stepperNum: int):
        if (stepperNum < 0) or (stepperNum >= _NUMBER_OF_DPi_STEPPER_DRIVERS):
            return False, 0
    
        results, velocityInStepsPerSec = self.getCurrentVelocityInStepsPerSecond(stepperNum)
        if results != True:
            return False, 0

        return True, velocityInStepsPerSec / self._stepsPerRevolution[stepperNum]


    #
    # home the motor by moving until the homing sensor is activated, then set the 
    # position to zero, with units in Revolutions
    #  Enter:  stepperNum = stepper driver number (0 - 2)
    #          directionTowardHome = 1 to move in a positive direction, -1 to move in 
    #            a negative directions 
    #          speedInRevolutionsPerSecond = speed to accelerate up to while moving 
    #            toward home, units in revolutions/second
    #          maxDistanceToMoveInRevolutions = unsigned maximum distance to move 
    #            toward home before giving up
    #  Exit:   True returned on success, else False
    #
    def moveToHomeInRevolutions(self, stepperNum: int, directionTowardHome: int, 
       speedInRevolutionsPerSecond: float, maxDistanceToMoveInRevolutions: float):
        
        return self.moveToHomeInSteps(stepperNum, directionTowardHome,
                                       int(speedInRevolutionsPerSecond * self._stepsPerRevolution[stepperNum]),
                                       int (maxDistanceToMoveInRevolutions * self._stepsPerRevolution[stepperNum]))

